@@ -64,6 +64,9 @@ import time
 from functools import partial as p
 from io import StringIO
 
+import pycond
+from devapp.tools import parse_kw_str
+
 env = os.environ
 wait = time.sleep
 now = time.time
@@ -82,14 +85,32 @@ mk_console = '''
 
 
 def check_assert(ass, res):
+    msg = 'Assertion failed: Expected "%s" not found in result (\n%s)'
     if ass is None:
         return
     s = str(res)
+    is_pycond = False
+    if isinstance(ass, str) and ' and ' in ass or ' or ' in ass:
+        is_pycond = True
+    elif isinstance(ass, list) and ass[1] in {'and', 'or'}:
+        is_pycond = True
+    if is_pycond:
+
+        def f(k, v, state, **kw):
+            return (k in state['res'], v)
+
+        r = pycond.pycond(ass, f)
+        a = r(state={'res': s})
+        if not a:
+            raise Exception(msg % (ass, s))
+        return
+
     if not isinstance(ass, (list, tuple)):
         ass = [ass]
+
     for a in ass:
+
         if not a in s:
-            msg = 'Assertion failed: Expected string "%s" not found in result (\n%s)'
             raise Exception(msg % (a, s))
 
 
@@ -213,7 +234,10 @@ def spresc(cmd):
 
 def sprun(*a, **kw):
     """Running a command as subprocess"""
-    print('--> ', a, kw)  # for the user (also in view messages)
+    if not kw and len(a) == 1:
+        print('', a[0])
+    else:
+        print('', a, kw)  # for the user (also in view messages)
     return p(sp.check_output, shell=True, stderr=sp.STDOUT)(*a, **kw)
 
 
@@ -366,6 +390,51 @@ def init_prompt(n):
     sprun("tmux send-keys -t %s:1 '' Enter" % n)
 
 
+def check_inline_lp(cmd, fn_lp):
+    if not isinstance(cmd, str):
+        return
+    l = cmd.rsplit(' # lp: ', 1)
+    if len(l) == 1 or '\n' in l[1]:
+        return
+    err, res = parse_header_args(l[1], fn_lp=fn_lp)
+    if err:
+        raise Exception(
+            'Inline lp construct wrong: %s. Valid e.g.: "ls -lta /etc # lp: expect=hosts timeout=10". Got: %s %s'
+            % (cmd, res[0], res[1])
+        )
+    res[1]['cmd'] = res[1].get('cmd', l[0])
+    return res[1]
+
+
+def parse_header_args(header, **ctx):
+    """Parsing either python or easy format
+    CAUTION: This is used in doc.pre_process as well, to parse markdown headers
+    """
+    ctx['dir_repo'] = ctx['fn_lp'].split('/docs/', 1)[0]
+    ctx['get_args'] = get_args
+
+    # the only function from devapp. For standalone you'd need to supply it:
+    from devapp.tools import parse_kw_str
+
+    try:
+        if header.rstrip().endswith('"'):
+            raise Exception('Not tried (apostrophe at end)')
+        return 0, ((), parse_kw_str(header, ctx, try_json=False))
+    except Exception as ex:
+        ex1 = ex
+        # evaling now.
+        # still - we supply only a minimum eval ctx and prevent
+        # imports. But still this won't be totally safe.
+        # BUT: Hey - we are about to run code *anyway*, this is LP in the end!
+        if not 'import' in header:
+            try:
+                m = eval('get_args(%s)' % header, ctx, {})
+                return 0, (m['args'], m['kw'])
+            except Exception as ex:
+                ex2 = ex
+        return 1, (ex2, ex1)
+
+
 class session:
     def get(session_name, **kw):
         """
@@ -378,8 +447,13 @@ class session:
             s = session_name
             # path is set new. bash (if executing user's shell is fish we'd be screwed)
             sprun('export SHELL=/bin/bash; export p="$PATH"; tmux new -s %s -d' % s)
-            a = 'tmux send-keys -t %(session)s:1 \'export PATH="$p"; export PS1="%(prompt)s" \' Enter'
-            b = {'prompt': kw.get('prompt', '$ '), 'session': s}
+            dt = []
+            for k in [i for i in env if i[:3] == 'DT_']:
+                dt.append('%s="%s"' % (k, env[k]))
+            dt = ' '.join(dt)
+            a = 'tmux send-keys -t %(session)s:1 \'export PATH="$p" PS1="%(prompt)s" %(dt)s\' Enter'
+
+            b = {'prompt': kw.get('prompt', '$ '), 'session': s, 'dt': dt}
             for i in (1, 2):
                 try:
                     sprun(a % b)
@@ -403,6 +477,7 @@ class session:
                             'configured with base index 1? 0 is default but will NOT work!!'
                         )
                     )
+
             init_prompt(s)
             if kw.get('root'):
                 sprun('tmux send-keys -t %s "sudo bash" Enter' % s)
@@ -428,6 +503,10 @@ class session:
         # TODO: clean up
         assert_ = None
         silent = kw.get('silent')
+
+        c = check_inline_lp(cmd, fn_lp=kw.get('fn_doc'))
+        if c:
+            cmd = c
         if isinstance(cmd, dict):
             timeout = cmd.get('timeout', timeout)
             expect = cmd.get('expect', expect)
@@ -444,9 +523,12 @@ class session:
             return 'silent'
 
         expect_echo_out_cmd = ''
+        is_multiline = '\n' in cmd
         if expect is None:
+            sep = '\n' if is_multiline else ';'
+            # here docs can't have ';echo -n... at last line:
             # not match on the issuing cmd
-            expect_echo_out_cmd = ';echo -n ax_; echo -n done'
+            expect_echo_out_cmd = sep + 'echo -n ax_; echo -n done'
             cmd += expect_echo_out_cmd
             expect = 'ax_done'
         if expect is False:
@@ -455,7 +537,18 @@ class session:
             expectb = expect.encode('utf-8')
         if cmd:
             init_prompt(n)
-            spresc("tmux send-keys -t %s:1 '%s' Enter" % (n, cmd))
+            # send the sequence as hex (-H):
+            seq = ' '.join([hex(ord(b))[2:] for b in cmd])
+            seq += ' a'
+            # if is_multiline:
+            #     breakpoint()  # FIXME BREAKPOINT
+            #     for line in cmd.splitlines():
+            #         spresc("tmux send-keys -t %s:1 '%s' Enter" % (n, line))
+            # else:
+            #     spresc("tmux send-keys -t %s:1 '%s' Enter" % (n, cmd))
+            print(' cmd to tmux: ', cmd)
+            spresc('tmux send-keys -t %s:1 -H %s' % (n, seq))
+
         t0 = now()
         wait_dt = 0.1
         while True:
@@ -477,7 +570,9 @@ class session:
             # when expect was given we include it (expect="Ready to accept Connections")
             # expect_echo_out_cmd is empty then
             res = res.split(expect, 1)[0].strip()
-            res = res.replace(expect_echo_out_cmd, '')
+            a = expect_echo_out_cmd
+            a = a[1:] if a.startswith('\n') else a  # when \n is the sep we won't see it
+            res = res.replace(a, '')
         else:
             # the tmux window contains a lot of white space after the last output when short cmd
             res = res.strip()
@@ -623,17 +718,36 @@ def repl_dollar_var_with_env_vals(kw, *keys):
 
 
 def multi_line_to_list(cmd):
+    """
+
+    ```bash lp session=DO asserts=loadbalancer.tf
+    ip () { echo 1.2.3.4; }
+    cat << FOO > loadbalancer.tf
+    > resource "digitalocean_loadbalancer" "www-lb" {
+    >     ipv4 = $(ip)
+    >     (...)
+    >}
+    >FOO
+    ls -l
+    ```
+
+    """
     if not isinstance(cmd, str):
         return cmd
-    try:
-        l = cmd.split('\n')
-    except Exception as ex:
-        print('breakpoint set')
-        breakpoint()
-        keep_ctx = True
-    if len(l) == len([i for i in l if not i.startswith(' ')]):
-        return l
-    return cmd
+    lines = cmd.split('\n')
+    # when some lines start with ' ' we send the whole cmd as one string:
+    if any([l for l in lines if l.strip() and l.startswith(' ')]):
+        # echo 'foo
+        #      bar' > baz
+        return cmd
+    r = []
+    while lines:
+        l = lines.pop(0)
+        r.append(l)
+        while lines and lines[0].startswith('> '):
+            l = lines.pop(0)
+            r[-1] += '\n' + l[2:]
+    return [cmd for cmd in r if cmd.strip()]
 
 
 def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
@@ -643,6 +757,7 @@ def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
     fn_doc: required: location of source file (async flow links contain its name)
     """
     # in python sigature format assert would be forbidden, so we allow asserts=...
+
     assert_ = kw.get('asserts') or kw.get('assert')
     repl_dollar_var_with_env_vals(kw, 'fn', 'cwd')
     # you could set this to org:
@@ -659,7 +774,11 @@ def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
     cwd = kw.get('cwd')
     if cwd:
         here = os.getcwd()
-        os.chdir(cwd)
+        try:
+            os.chdir(cwd)
+        except Exception as ex:
+            ex.args += ('dest dir: "%s"' % cwd,)
+            raise
 
     # mode = kw.get('xmode')
     # # in python we need tmux session (to start python first):
@@ -679,6 +798,9 @@ def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
 
     session_name = kw.pop('session', 0)
     if session_name:
+        if kw.get('mode'):
+            msg = 'modes are not supported with session (got mode=%(mode)s)'
+            raise Exception(msg % kw)
         res = session.srun(cmd, session_name=session_name, **kw)
         if cwd:
             os.chdir(here)
@@ -702,6 +824,7 @@ def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
 
         res = []
         for c in cmd:
+            c = check_inline_lp(c, fn_lp=fn_doc) or c
             c1 = c['cmd'] if isinstance(c, dict) else c
             rcmd = c1
             prompt = '$'
@@ -715,10 +838,13 @@ def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
             # return sp.check_output(['sudo podman ps -a'], shell=True)
             r = sprun(rcmd)
             r = r.decode('utf-8').rstrip()
+            if isinstance(c, dict):
+                check_assert(c.get('assert', c.get('asserts')), r)
             r = ''.join([prompt, ' ', c1, '\n', r])
             # res = '%s %s\n' % (prompt, cmd) + res
 
             res.append({'cmd': c1, 'res': r})
+
     check_assert(assert_, res)
     # cwd header only for the current block:
     if cwd:
@@ -734,9 +860,49 @@ def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
     # if now() - t0 > dt_cache:
     cache.add(cmd, res, kw)
     last_result[0] = res
+
+    i = int(kw.get('addsrc', 0))
+    if i:
+        b = '\n' + kw.get('sourceblock', 'n.a.')
+        m = {
+            'blocksource1': b.replace('\n', '\n '),
+            'blocksource4': b.replace('\n', '\n    '),
+            'res': res,
+            'res4': ('\n' + res).replace('\n', '\n    '),
+        }
+        f = getattr(BlockSrc, 'fmt_%s' % i, BlockSrc.fmt_1)
+        res = f % m
     return res
 
 
+class BlockSrc:
+    fmt_1 = '''
+
+LP Source:
+
+```bash
+%(blocksource1)s
+```
+
+Result: 
+
+%(res)s
+'''
+
+    fmt_2 = '''
+
+=== "LP Source"
+
+    ```bash
+    %(blocksource4)s
+    ```
+
+=== "Result" 
+
+    %(res4)s
+'''
+
+
 def test():
     """Better use docutools lp_client to test"""
     print(
