@@ -131,7 +131,15 @@ class Flags:
         d = False
 
     class lit_prog_skip_existing:
-        n = 'Intended for CI: When seconary .md pages had been committed, e.g. built while authoring, CI will not re-evaluate those pages at presence of that flag'
+        """Intended for CI or to prevent unwanted re-evaluation in general:
+        When secondary .md pages had been committed, i.e. are present on the filesystem
+        (built while authoring), we will NOT re-evaluate those pages at presence of that
+        flag.
+        In order to get on-demand eval, just run lp w/o that flag and possibly a match
+        on your source file (-lpe=mysourcefile instead -lpe=md). Or delete the md.
+        """
+
+        n = 'Only eval when there are no .md files'
         d = False
 
     class lit_prog_on_err_keep_running:
@@ -263,6 +271,11 @@ class Credits:
             indirect_dependencies -= direct_dependencies
             indirect_dependencies -= dev_dependencies
             dependencies = direct_dependencies | dev_dependencies | indirect_dependencies
+            app.info(
+                'Generating credits',
+                dependencies=dependencies,
+                indirect_dependencies=indirect_dependencies,
+            )
 
             packages = {}
             attrs = ('name', 'home-page', 'license', 'version', 'summary')
@@ -275,7 +288,7 @@ class Credits:
                     pkg = Credits.fetch_pypi(dependency, attrs)
                     packages[pkg['name'].lower()] = pkg
                 else:
-                    app.info('found', dependency=dependency)
+                    app.debug('found', dependency=dependency)
 
             for d in indirect_dependencies:
                 if not d in packages:
@@ -864,6 +877,9 @@ class LP:
         raise_on_errs intended for temporarily changing behviour, e.g. for tests
         Else use the FLG.
         """
+        if os.path.exists(fn_lp):
+            os.environ['DT_DOCU'] = os.path.dirname(fn_lp)
+            os.environ['DT_DOCU_FILE'] = fn_lp
         S.cur_fn_lp = fn_lp
         lp_blocks, dest = do(LP.extract_lp_blocks, md=md, fn_lp=fn_lp)
         app.info('---------- %s --------------' % fn_lp)
@@ -903,10 +919,19 @@ class LP:
             if (cmd.strip() + ' ')[0] in ('[', '{'):
                 try:
                     cmd = literal_eval(cmd)
-                except:
-                    cmd = json.loads(cmd)
+                except Exception as exle:
+                    try:
+                        cmd = json.loads(cmd)
+                    except Exception as ex:
+                        ex.args += ('LP: Expression to deserialize was: %s' % cmd,)
+                        ex.args += (
+                            'LP: Before json.loads we tried literal eval but got Exception: %s'
+                            % exle,
+                        )
+                        raise
             # cmd = block['code']
             kw['lang'] = block.get('lang')
+            kw['sourceblock'] = block.get('source')
             S.lp_stepmode and LP.confirm('Before running', page=fnd, cmd=cmd, **kw)
             run_lp = partial(lit_prog.run, fn_doc=fnd)
             kw['timeout'] = kw.get('timeout', S.lp_evaluation_timeout)
@@ -949,32 +974,14 @@ class LP:
 
     def extract_header_args(lp_header, fn_lp):
         H = ' '.join(lp_header.split()[2:])
-        header_kws = LP.header_kws(fn_lp)
+        err, res = lit_prog.parse_header_args(H, fn_lp=fn_lp, dir_project=project.root())
+        if not err:
+            return res
 
-        try:
-            return (), parse_kw_str(H, header_kws, try_json=False)
-        except Exception as ex:
-            ex1 = ex
-            # evaling now.
-            # still - we supply only a minimum eval ctx and prevent
-            # imports. But still this won't be totally safe.
-            # BUT: Hey - we are about to run code *anyway*, this is LP in the end!
-            if not 'import' in H:
-                try:
-                    return eval('get_args(%s)' % H, header_kws, {})
-                except Exception as ex:
-                    ex2 = ex
-        return LP.header_parse_err, {LP.py_err: ex2, LP.easy_args_err: ex1, 'header': H}
-
-    get_args = lambda *a, **kw: [a, kw]  # trick to get python sig args
-
-    def header_kws(fn_lp):
-        """constants you can use in headers"""
-        return {
-            'get_args': LP.get_args,
-            'dir_repo': fn_lp.split('/docs/', 1)[0],
-            'dir_project': project.root(),
-        }
+        return (
+            LP.header_parse_err,
+            {LP.py_err: res[0], LP.easy_args_err: res[1], 'header': H},
+        )
 
     def extract_lp_blocks(md, fn_lp):
         s = md.splitlines()
@@ -1007,10 +1014,12 @@ class LP:
                     break
             if add:
                 continue
+            src_header = dest[-1].strip()
             lpnr += 1
             lp_header = dest.pop(-1)
             dest.append(LP.PH(lpnr))  # placeholder
             n = [l[len(ind) :] for l in n[:-1]]
+            source = src_header + '\n' + '\n'.join(n) + '\n```'
             l = fragm[0].split('```', 1)[1].strip()
             if not FLG.lit_prog_debug_matching_blocks in lp_header:
                 continue
@@ -1022,18 +1031,31 @@ class LP:
                 'args': a,
                 'kwargs': kw,
                 'indent': ind,
+                'source': source,
             }
 
             lps.append(spec)
         return lps, dest
 
     def is_lp(d, fn, match=''):
+        """
+        We get all files in the docs dir matching match (md)
+
+        """
         # and fm in fn
-        # and not fn.endswith('.lp.md')Arbeitsweise
+        # and not fn.endswith('.lp.md')
         # and LP.fn_lp(fn) in mkdocs
-        if S.lp_evaluation_skip_existing and exists(fn.rsplit('.lp', 1)[0]):
+        if not fn.endswith('.md.lp'):
             return
-        return fn.endswith('.md.lp') and match in (d + '/' + fn)
+        r = ''
+        if S.lp_evaluation_skip_existing and exists(d + '/' + fn.rsplit('.lp', 1)[0]):
+            r = '.md exists, skip_existing set'
+        if not match in (d + '/' + fn):
+            r = 'Not matching %s' % match
+        if r:
+            app.info('LP: skipping %s' % fn, reason=r)
+            return
+        return True
 
     def verify_no_errors(files):
         errs = []
@@ -1054,6 +1076,7 @@ class LP:
         os.makedirs(project.root() + '/tmp/tmux', exist_ok=True)
         files = walk_dir(docs, crit=partial(LP.is_lp, match=file_match))
         do_files = []
+        os.environ['DT_PROJECT_ROOT'] = project.root()
         for fn in files:
             t = os.stat(fn)[8]
             if t == S.lp_files.get(fn):
